
# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2012, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

module Tomte
autoload :Logger,           'tomte/logger'

module Agent

  autoload :Daemon,         'tomte/agent/daemon'
  autoload :Configuration,  'tomte/agent/configuration'

  #
  # Provides the agent skeleton, common to both sync and async agents
  #
  class Common
    extend Tomte::Agent::Configuration

    attr_accessor :log

    #
    # gives you a new agent of the defined type (Agent::Sync, Agent::Async)
    #
    def initialize(options = {})

      config.core[:argv] = options[:argv] if  options[:argv]

      Tomte::Agents.configure! :parse_command_line => true

      #Dir.chdir(config.agents[:chdir]) if config.agents[:chdir]

      #  we do start the logger here
      @log = options[:logger] || Tomte.logger
      @log.debug "+ #{self.class} initializing."
      @log.debug "- Configuration \n" + config.debug if debug?

      @exit_code = 0
      @signals_queue = []

      if config.agents[:privilege_management]
        pre_init(options)
        change_privileges
      end
    end

    #
    # called to effectively run the agent. Including demonizing, if requested
    #
    def run!(options = {})

      init options

       # daemonize
      daemon? ? daemonize { start_agent(options) } : start_agent(options)
      self

    rescue Exception => e
      stray_exception e
    end

    #
    #
    #
    def run_agent
      config.core[:command] ? run_command(config.core[:command], config.core[:arguments]) : run
      terminate
    end

    #
    # runs a specific command with this agent
    #
    def run_command(command, args)
      command = command.to_sym

      if self.respond_to? command, true
        send(command, *args)

      else
        log.error("= Unknown command #{command}")
        shutdown!
      end

    rescue ArgumentError => e
      log.error("Argument Error for method #{command}:" + e.to_s)
      tomte_debug { log.exception e }
      shutdown!
    end

    #
    # usually called to quit agent
    #
    def shutdown!(options = {})
      after_shutdown(options)
    end

    #
    # sets an exit code for this agent
    #
    def set_exit_code(code)
      @exit_code = code
    end

    #
    # True if this agent is (or should be) a daemon
    #
    def daemon?
      config.agents[:daemonize]
    end

    #
    # true if we are configurated to run in debugging mode
    #
    def debug?
      config.core[:debug]
    end


    private
    #######
    #

    #
    # sets up signal traps in a 2.0 compatible fashion
    #
    def setup_signal_traps
      %w(INT TERM USR1 USR2).each do |sig|
        trap sig do
          @signals_queue << sig
         end
       end
    end

    #
    # polls the signal queue for pending signals
    #
    def handle_signals
      while sig = @signals_queue.shift
        log.debug "! Received #{sig} signal."
        case sig
          when 'INT', 'TERM'
            shutdown!
        end
      end
    end

    def after_shutdown(options = {})
      log.debug "= Quitting."
      daemon_shutdown if daemon?
      terminate
      exit(options[:exit_code] || @exit_code || 0)
    end

    def stray_exception(e)
      raise e if e.is_a? SystemExit

      if log
        log.error "= ====== unhandled exception [#{e.class.name}] ======".important!
        log.exception(e)
      else
        raise e
      end

      # do something better than just shutdown? XXX
      shutdown!
    end

    def change_privileges

      user = config.agents[:user]  || return
      group = config.agents[:group] || 'tomte'


      log.info "+ Changing process privilege to #{user}:#{group}"

      uid, gid = Process.euid, Process.egid
      target_uid = Etc.getpwnam(user).uid
      target_gid = Etc.getgrnam(group).gid

      if uid != target_uid || gid != target_gid

        Process.initgroups(user, target_gid)
        Process::GID.change_privilege(target_gid)
        Process::UID.change_privilege(target_uid)
      end
    rescue Errno::EPERM => e
      log.error "! Could not change user and group to #{user}:#{group}: #{e}"
      shutdown!
    end

    #
    #
    # called before dropping privileges
    #
    def pre_init(options = {})

    end

    #
    # performs the agent initialization,
    # called after the internal inits, at the beginning of the run! cycle
    #
    def init(options = {})
      log.debug "= #{self.class} has empty init"
    end


    #
    # called after execution has ended
    #
    def terminate

    end

    #
    # accessor to the config singleton
    #
    def config
      Tomte.config
    end

    def identity
      config[:local][:identity] + '@' + config[:local][:host]
    end

  end
end
end
