
# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2012, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

module Tomte
  class Director < Tomte::Actor
  include Tomte::Protocol::Async::Callbacks

    class Wall

      attr_accessor :dispatcher

      Levels = [:fatal, :error, :warn, :info].freeze

      Levels.each do |level|
        define_method(level) do | caption, attributes = {} |
          publish(level, caption, attributes)
        end
      end

      private
      def publish(level, caption, attributes = {})
        return unless @dispatcher
        @dispatcher << [:wall, level, caption, attributes]
      end
    end

    DEFAULT_DISPATCHER = 'Tomte::Workers::SimpleDispatcher'

    #
    # This director log instance
    #
    attr_accessor :log

    #
    # The wall for publishing network alerts
    #
    attr_reader :wall

    #
    # The registry takes care of managing workers
    #
    attr_reader :registry

    #
    # The dispatcher takes care of managing messages to the management bus
    #
    attr_reader :dispatcher

    #
    # the director identity is also the host identity
    #
    attr_reader :uuid, :identity, :host

    #
    # reference to the system configuration
    #
    attr_reader :config


    def initialize(options = {})
      super()

      initialize_callbacks

      @config = Tomte::Config

      @uuid     = options.fetch(:uuid,        config.local[:uuid])
      @identity = options.fetch(:identity,    config.local[:identity])
      @host     = options.fetch(:host,        config.local[:host])
      @dispatcher_class =
                  options.fetch(:@dispatcher_class, DEFAULT_DISPATCHER)

      @log = options[:logger] || Tomte::Logger.new(options)
      @log.info '+ Identity is: ' + "#{@identity}@#{@host}".blue
      @log.info '+ UUID is: ' + "#{@uuid}".yellow
      @exiting = false

      @wall = Wall.new

      append_callback :after_shutdown, &options[:after_shutdown] if
          (options[:after_shutdown] && options[:after_shutdown].respond_to?(:call))

      events
      init(options)
    end

    #
    # Init is called when the director is actually set and ready to process events
    #
    def init(options = {})
      log.debug "+ Director Started."
      @registry   = spawn_worker('Tomte::Workers::Registry', :ident => :registry, :name => 'system registry')
      @dispatcher = spawn_worker(@dispatcher_class, :ident => :dispatcher, :name => 'system dispatcher')
      @wall.dispatcher = @dispatcher

      start_workers :system_workers
    end


#
    # register events
    #
    def events
      receive do |filter|

        #
        # registry is ready
        #
        filter.when :registry_ready do

        end

        #
        # dispatcher is ready to perform its functions
        #
        filter.when :dispatcher_ready do

          # alright sign-on the supervision network
          dispatcher << :sign_on
        end

        #
        # received when registry has finished shutting down
        #
        filter.when :registry_exited do
          if dispatcher
            dispatcher << :shutdown
          else
            exec_callback_once :after_shutdown
          end
        end

        #
        # dispatcher has exited
        #
        filter.when :dispatcher_exited do

          log.debug "- Network processes have exited, closing."
          exec_callback_once :after_shutdown
        end

        #
        # sent by the dispatcher when supervision request failed
        #
        filter.when :unsupervised_trying do

          log.warn "+ You have requested supervision but we're still waiting for it"
        end

        #
        # sent by the dispatcher when this agent is successfully supervised, so we're ready to start
        #
        filter.when :network_ready do

          log.debug "+ Network supervision ready."
          ready
        end


        #
        # another workers asks the director to shut itself down after closing the whole system
        #
        filter.when :shutdown do
          shutdown
        end


        #
        # Another worker asks to spawn a new worker
        #
        filter.when(T[:spawn_worker, String, Hash]) do |_, klass, params|
          spawn_worker(klass, params)
        end
      end
    end



    #
    # ask the director to shut itself down, after having closed all the workers
    #
    def shutdown(&block)
      return if exiting?

      @exiting = true
      log.debug "- Director asked to shutdown."

      append_callback :after_shutdown, &block if block_given?
      exec_callback :before_shutdown

      # ask the registry to notify all the workers they should shut down
      registry << :shutdown if registry
    end

    #
    # Spawn an instance of a worker and register it
    #
    def spawn_worker(klass, params= {})
      begin
        Tomte::Agents::Framework.preload_worker(klass, params)
        new = klass.constantize.new(self, params)
        registry.register(new, params) unless registry.nil?

      rescue Exception => e
        log.exception e
      end
      new
    end


    def exiting?
      @exiting
    end


    private
    #######


    def martian(m)
      log.warn "+ Got a martian #{m.inspect}"
    end


    def ready
      start_workers
    end

    def start_workers(list = :workers)
      return if (@workers_started ||= {})[list]

      workers = config.agents[list] || {}
      workers.each do |label, worker|
        log.debug "- Director starting worker '#{worker}'"
        raise ArgumentError, "missing a name or a class for the requested worker" if worker[:class].blank? or label.blank?
        worker[:name] ||= label
        self << [:spawn_worker, worker[:class], worker]
      end
      @workers_started[list] = true
    end

  end
end
