
# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2012, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

module Tomte

  autoload :Actor,      'tomte/async/actor'
  autoload :Director,   'tomte/async/director'


  class Worker < Actor
    extend Forwardable

    include Tomte::Protocol::Async::Callbacks
    include Tomte::Network::Bus
    include Tomte::Network::API
    include Tomte::Network::Debugger
    include Tomte::Network::Mgmt::Methods

    # quick reference to the Director
    attr_reader :director

    # handy reference to the log
    attr_accessor :log

    # identity for this worker
    attr_reader :identity

    #
    attr_reader :uuid


    # name for this worker
    attr_reader :name

    # when this worker was started
    attr_accessor :started

    # is this a system worker?
    class_attribute :system_worker

    # this worker shouldn't be put in the registry
    class_attribute :do_not_registry

    def initialize(director=nil, params = {})
      super()

      initialize_callbacks

      @name =     params.fetch(:name,      "unnamed #{self.class}")
      @identity = params.fetch(:identity,  "#{@name}@#{config.local[:identity]}")
      @uuid =     params.fetch(:uuid,      UUIDTools::UUID.random_create.to_s)

      @started = Time.now
      @exiting = false

      @director = director
      @endpoints =  {}
      @log = params[:log] || if (director && director.respond_to?(:log))
                               director.log
                             else
                               Tomte::Logger.new(STDOUT)
                             end

      events

      append_callback :all_endpoints_closed, &method(:done_shutdown)
      append_callback(:all_endpoints_ready) {
        log.debug "+ #{self.class.name}: all endpoints ready."
        init(params)
      }

      initialize_endpoints
      initialize_apis
      initialize_remote_debugger unless system_worker?
    end


    def init(options = {})

    end

    #
    # true if this worker is in the exit process
    #
    def exiting?
      @exiting
    end

    def config
      Tomte.config
    end

    def self.config
      Tomte.config
    end

    private
    #######

    def events
      receive do |filter|

        # received shutdown event
        filter.when :shutdown do
          shutdown
        end

      end
    end

    #
    # shuts down the system, by asking the director
    #
    def shutdown!
      log.debug "- Worker '#{name}:#{identity}' is shutting down the system."
      director << :shutdown
    end

    #
    # shuts down this agent
    #
    def shutdown
      return if exiting?

      @exiting = true

      log.debug "- Shutting down this worker '#{name}:#{identity}'."
      exec_callback :before_shutdown
    end

    def done_shutdown
      exec_callback_once :after_shutdown
      director.registry << [:worker_has_quit, identity]
      log.debug "- Agent: '#{name}:#{identity}' has quit. Goodbye."
    end

    class << self

      #
      # sets parameters for the worker
      #
      def worker(options = {})
        self.system_worker = options.fetch(:system, false)
        self.do_not_registry = options.fetch(:do_not_registry, false)
      end

    end
  end
end
