
# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2012, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

require 'statemachine'
require 'tomte/async/worker'

module Tomte
  class Dispatcher < Tomte::Worker

    DEFAULT_SUPERVISOR_INTERFACE = 'tomte.supervisor.agent'
    DEFAULT_EVENT_EXCHANGE = 'tomte.event'
    DEFAULT_SIGNON_RETRY = 30 ### for testing better like 600
    DEFAULT_HEARTBEAT_TIME = 30
    HEARTBEAT_ENABLE = true
    EVENTS_ARE_RPC = true

    worker :system => true,
           :bus => :management


    class SupervisorInterface
      extend Forwardable

      class Info < Tomte::Message::Data; end

      # capabilities
      # :hb
      # :sync| :asyn
      attr_accessor :info, :states
      def_delegators :@states, :state

      #
      # gets you a supervisor interface
      #
      def initialize(client)
        @client = client

        @info = Info.new
        @info.capabilities = [:hb, :_async]
        @info = @info.merge Tomte::Message.local_info

        define_states
        yield info if block_given?
      end


      #
      #
      #
      def hello
        new_message(:hello)
      end


      private
      #######

      def new_message(event)
        Tomte::MM::Message.new do |m|
          m.header.event = event
          m.header.key = 'event.presence.' + event.to_s
        end
      end

      def define_states
        @states = Statemachine.build do
          state :idle do
            event :start_unsupervised, :unsupervised_ready
            event :request_sign_on, :requesting_sign_on, :request_sign_on
            event :sign_on_failed, :unsupervised_trying
          end
          state :requesting_sign_on do
            event :sign_on_barred, :exited
            event :sign_on_ok, :ready
            event :sign_on_failed, :unsupervised_trying
            event :request_sign_off, :exited ## hmmm?
          end
          state :unsupervised_ready do
            on_entry :ready
            event  :request_sign_off, :exited # maybe the signal name should be changed? XXX FIXME
          end
          state :ready do
            on_entry :ready
            event :request_sign_off, :requesting_sign_off, :request_sign_off
          end
          state :unsupervised_trying do
            on_entry :unsupervised_trying
            event :request_sign_on, :requesting_sign_on, :request_sign_on
            event :request_sign_off, :exited
          end
          state :requesting_sign_off do
            event :sign_off_done, :exited
            event :request_sign_off, :requesting_sign_off
          end
          state :exited do
            on_entry :exited
          end
        end
        @states.context = @client
      end

    end
  end

  class Dispatcher < Tomte::Worker

    # exchange to public our events on to
    attr_accessor :event_exchange
    # the rpc where we send requests to the supervisor
    attr_accessor :supervisor_rpc, :sup
    # the rpc where we listen for requests
    attr_accessor :command_rpc

    attr_accessor :mq


    def init(options = {})
      super


      @supervisor_interface_name = config.core[:supervisor_interface] || DEFAULT_SUPERVISOR_INTERFACE
      @event_exchange_name = config.core[:event_exchange] || DEFAULT_EVENT_EXCHANGE


      setup_event_exchange
      setup_supervisor_rpc

    end

    #
    # register worker events
    #
    def events
      receive do |filter|

        filter.when :director_initialized do
          log.debug "- agent is initialized"
          config.agents[:unsupervised] ? event.start_unsupervised : event.request_sign_on
        end


        filter.when :shutdown do
          log.debug "- been requested to sign off and shut everything down"
          event.request_sign_off
        end


        # send an update message
        filter.when(:workers_update) do
          # dp nothing
        end

        #
        # a new worker has been started
        #
        filter.when(T[:notify_worker_start, Tomte::Worker]) do |_, worker|
          # do nothing
        end
      end
    end


    #
    # called when client is ready and authorized
    #
    def ready
      log.info '= agent is ready.'
      heartbeat_start
      director << :ready
    end


    def unsupervised_trying
      log.info "= agent is continuing to try a signon every #{DEFAULT_SIGNON_RETRY} seconds"
      EventMachine::Timer.new(DEFAULT_SIGNON_RETRY) do
        event.request_sign_on
      end
      director << :unsupervised_trying
    end

    # called when sign-off is ended
    def exited

      # close stuff?
      director << :dispatcher_exited
    end

    # state transitions
    #
    def request_sign_on
      begin
        log.debug "+ requesting sign-on to supervisor"
        @supervisor_rpc.agent_sign_on sup.info do |res|

          if (!res || res.is_a?(Tomte::RPC::Error))
            unless config.agents[:unsupervised]
              log.error "error signing on: '#{res.to_s}'"
            end
            event.sign_on_failed
          else

            case res.result
              when :ok
                log.debug '+ sign-on successful'
                event.sign_on_ok
              when :barred
                log.debug '+ sign-on barred'
                event.sign_on_barred
              else
                log.debug '+ sign on failed'
                event.sign_on_failed
            end
          end
        end
      rescue Exception => e
        log.error e
        event.sign_on_failed
      end
    end


    def request_sign_off
      @supervisor_rpc.agent_sign_off sup.info do |res|

         if (!res || res.is_a?(Tomte::RPC::Error))
          log.error 'error signing off'
          log.error res.to_s
          event.sign_off_done  ## done anyways
        else
          event.sign_off_done

        end
      end
    end


    def publish_workers_update
      if EVENTS_ARE_RPC
        rpc_publish { @supervisor_rpc.agent_worker_updates sup.info.local.uuid, director.registry.workers_list }
      else

      end
    end

    def publish_hello
      if EVENTS_ARE_RPC
        rpc_publish { @supervisor_rpc.agent_hello sup.info.local.uuid }
      else
        publish sup.hello
      end
    end

    private
    #######

    def event
      sup.states
    end

    # publish only if we have a supervisor
    def rpc_publish
      return unless @sup.state == :ready
      yield
    end

    def heartbeat_start
      #
      # should send an agent presence message
      #

      @hello_timer = EventMachine::PeriodicTimer.new(@hb_time) do
        publish_hello
      end
    end


    def publish(message)

      log.debug "- dispatcher sending message '#{message.key}'"
      log.debug message.inspect
      begin
        event_exchange.publish(message, :key => message.key)
      rescue AMQP::Error => e
        log.error "= dispatcher: " + e.message
      end
    end

    def setup_event_exchange
      log.debug "- Dispatcher setting up event exchange on '#{@event_exchange_name}'"
      @mq = endpoint.new_channel
      @event_exchange = mq.topic(@event_exchange_name, :durable => true)
      @hb_time = config.agents[:heartbeat] || DEFAULT_HEARTBEAT_TIME
      @hb_enable = HEARTBEAT_ENABLE
    end

    def setup_supervisor_rpc
      log.debug "- Dispatcher setting up supervisor interface on '#{@supervisor_interface_name}'"
      @supervisor_rpc = endpoint.xrpc(:queue => @supervisor_interface_name, :timeout => 5)
      @sup = SupervisorInterface.new(self) do |i|
        i.capabilities = [:_async]
        i.capabilities << :hb if @hb_enable
        i.hb_time = @hb_time
      end
      setup_return_callback
    end

    def setup_return_callback
       ## registers the returned handler
      @supervisor_rpc.__mq.return_callback do |reason, header, info|
        unless config.agents[:unsupervised]
          log.error "! Eep! Bus returned message because of error #{reason[:code]}:#{reason[:text]}"
          log.debug header.inspect
          log.debug info
        end
      end
    end

  end
end
