
# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2012, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

module Tomte::Workers
  class Registry < Tomte::Worker

    FORCE_SHUTDOWN_DELAY = 10

    attr_accessor :workers
    worker :system => true, :do_not_registry => true

    def init(params ={})

      # Registry is a special worker that doesn't want to close
      # after all endpoints are closed
      clear_callbacks :all_endpoints_closed

      @workers = {}
      @name = 'system registry'
      @identity = 'registry'
      @exiting = false

      register(self)
      director << :registry_ready
    end


    def events
      receive do |filter|

        filter.when :shutdown do
          shutdown
        end

        filter.when(T[:worker_has_quit, Object]) do |_, worker|
          de_register(worker)
        end

        filter.when(T[:register, Tomte::Worker]) do |_, worker|
          register(worker)
        end

      end
    end


    #
    # registers a new worker
    #
    def register(worker, params = {})
      return if worker.class.respond_to?(:do_not_registry?) && worker.class.do_not_registry?

      log.debug("+ Registry: asked to register #{worker.name}:#{worker.identity}")

      raise ArgumentError, 'asked to register a process that is not a worker' unless worker.is_a? Tomte::Actor

      if workers.has_key?(worker.identity)
        raise ArgumentError, 'duplicate worker identity'
      end

      workers[worker.identity] = worker
      notify_dispatcher
    end


    #
    # de-register a worker by identity or object
    #
    def de_register(worker)

      identity = (worker.respond_to?(:identity) ? worker.identity : worker)

      log.debug("- Registry: de-registering '#{identity}'.")
      @workers.delete(identity)

      done_shutdown if(exiting? && @workers.empty?)
    end

    #
    # return worker with this identity
    #
    def [](identity)
      @workers[identity]
    end

    #
    # returns a list of the workers in this registry
    #
    def workers_list
      workers.map do |k, w|
        {:uuid => w.uuid, :identity => w.identity, :type => w.class.name, :name => w.name, :started => w.started}
      end
    end

    #
    # tell a worker to shut itself down
    #
    def shutdown_worker(identity)

      log.debug("- Registry: requiring shutdown of '#{identity}'.")
      @workers[identity] << :shutdown
    end


    #
    # tell all the workers to shut themselves down
    #
    def shutdown_all_workers
      @workers.each {|identity, _| shutdown_worker(identity)}
    end


    #
    # shutdown this registry
    #
    def shutdown
      return if exiting?

      log.debug("- Registry: shutting down.")

      # shutdown anyways after a while
      Tomte::Timer.delay(FORCE_SHUTDOWN_DELAY) { done_shutdown }

      super
      shutdown_all_workers
    end


    #
    # sends the dispatcher an update
    #
    def notify_dispatcher
      director.dispatcher << :workers_update if director.dispatcher
    end


    private
    #######

    def done_shutdown
      log.debug("- Registry: we have done, exiting.")
      @workers = {}
      @exiting = false
      director << :registry_exited
    end

  end
end
